use std::{
    collections::{BTreeMap, HashMap},
    fs::Metadata,
    future::ready,
    io::SeekFrom,
    path::{Path, PathBuf},
    pin::Pin,
    sync::Arc,
    task::{Context, Poll},
    time::{SystemTime, UNIX_EPOCH},
};

use axum::http::StatusCode;
use base64::{engine::general_purpose::STANDARD as base64std, Engine};
use bytes::{Buf, Bytes, BytesMut};
use futures::{future::BoxFuture, stream, Future, FutureExt, Stream, StreamExt, TryFutureExt};
use log::{error, info};
use sqlx::{query, MySqlPool};
use tokio::{
    fs::{metadata, read_dir, File, ReadDir},
    io::{AsyncReadExt, AsyncSeekExt, AsyncWriteExt},
    sync::RwLock,
};
use webdav_handler::{
    davpath::DavPath,
    fs::{
        DavDirEntry, DavFile, DavFileSystem, DavMetaData, DavProp, FsError, FsFuture, FsResult,
        FsStream, OpenOptions, ReadDirMeta,
    },
};
use xmltree::{Element, XMLNode};

use crate::{error::Error, fs::FsUser};

const DAV_OC_PROP: &str = "http://owncloud.org/ns";
const DAV_NC_PROP: &str = "http://nextcloud.org/ns";

#[derive(Clone)]
pub struct FilesFs {
    pub user: FsUser,
    pub db: MySqlPool,
    pub incoming_share_cache: Arc<RwLock<BTreeMap<PathBuf, (FsUser, PathBuf)>>>,
}

impl DavFileSystem for FilesFs {
    fn open<'a>(&'a self, path: &'a DavPath, options: OpenOptions) -> FsFuture<Box<dyn DavFile>> {
        info!("test: {}", path.to_string());
        Box::pin(async move {
            let (meta, path) = self
                .resolve_path(path)
                .map_err(|_| FsError::GeneralFailure)
                .await?;

            let mut tokio_options = tokio::fs::OpenOptions::new();
            tokio_options
                .read(options.read)
                .write(options.write)
                .append(options.append)
                .truncate(options.truncate)
                .create(options.create)
                .create_new(options.create_new);

            let file = OxiFile::new(path.clone(), &tokio_options).await?;
            let file: Box<dyn DavFile> = Box::new(file);

            Ok(file)
        })
    }

    fn read_dir<'a>(
        &'a self,
        ppath: &'a DavPath,
        meta: ReadDirMeta,
    ) -> FsFuture<FsStream<Box<dyn DavDirEntry>>> {
        info!("dir: {:?}", ppath.as_pathbuf());
        Box::pin(async move {
            //let (is_share, path) = self
            //    .resolve_path(ppath)
            //    .map_err(|_| FsError::GeneralFailure)
            //    .await?;
            //warn!("dirpath: {:?}", path);
            //if !self.can_access(&path) {
            //    return Err(FsError::Forbidden);
            //}

            let (meta, path) = self
                .resolve_path(ppath)
                .map_err(|_| FsError::GeneralFailure)
                .await?;

            let entry_stream = DirEntryStream::read_dir(&path).await.map(|x| {
                let y: Box<dyn DavDirEntry> = Box::new(x);
                y
            });

            let entry_stream = entry_stream.chain(stream::iter(self.shares(&ppath).await.unwrap()));
            let entry_stream: FsStream<Box<dyn DavDirEntry>> = Box::pin(entry_stream);

            Ok(entry_stream)
        })
    }

    fn metadata<'a>(&'a self, path: &'a DavPath) -> FsFuture<Box<dyn DavMetaData>> {
        Box::pin(async move {
            info!("metadata {:?}", path);
            let (meta, path) = self
                .resolve_path(path)
                .map_err(|_| FsError::GeneralFailure)
                .await?;

            if meta.exists {
                let file = OxiFile::new(path.clone(), tokio::fs::OpenOptions::new().read(true))
                    .await?
                    .metadata()
                    .await
                    .unwrap();

                Ok(file)
            } else {
                Err(FsError::NotFound)
            }
        })
    }

    fn symlink_metadata<'a>(&'a self, path: &'a DavPath) -> FsFuture<Box<dyn DavMetaData>> {
        // symlinks are currently not supported, so no difference to normal metadata
        self.metadata(path)
    }

    fn create_dir<'a>(&'a self, path: &'a DavPath) -> FsFuture<()> {
        Box::pin(async move {
            info!("create dir");
            let (path_meta, path) = self
                .resolve_path(path)
                .map_err(|_| FsError::GeneralFailure)
                .await?;

            match path_meta.exists {
                false => {
                    //TODO remove shares in db with src_path=path
                    tokio::fs::create_dir_all(path)
                        .await
                        .map_err(|_| FsError::GeneralFailure)?;
                    Ok(())
                }
                true => Err(FsError::Exists),
            }
        })
    }

    fn remove_dir<'a>(&'a self, path: &'a DavPath) -> FsFuture<()> {
        Box::pin(async move {
            info!("remove dir");
            let (path_meta, path) = self
                .resolve_path(path)
                .map_err(|_| FsError::GeneralFailure)
                .await?;

            match (path_meta.exists, path_meta.is_mount_point) {
                (true, false) => {
                    //TODO remove shares in db with src_path.starts_with(path)
                    tokio::fs::remove_dir(path)
                        .await
                        .map_err(|_| FsError::GeneralFailure)?;

                    Ok(())
                }
                (true, true) => Err(FsError::NotImplemented), //TODO remove share in db with dst_path =path
                _ => Err(FsError::NotFound),
            }
        })
    }

    fn remove_file<'a>(&'a self, path: &'a DavPath) -> FsFuture<()> {
        Box::pin(async move {
            info!("remove file");

            let (path_meta, path) = self
                .resolve_path(path)
                .map_err(|_| FsError::GeneralFailure)
                .await?;

            match (path_meta.exists, path_meta.is_mount_point) {
                (true, false) => {
                    //TODO remove shares in db with src_path=path
                    tokio::fs::remove_file(path)
                        .await
                        .map_err(|_| FsError::GeneralFailure)?;

                    Ok(())
                }
                (true, true) => Err(FsError::NotImplemented), //TODO remove share in db with dst_path =path
                _ => Err(FsError::NotFound),
            }
        })
    }

    fn rename<'a>(&'a self, from: &'a DavPath, to: &'a DavPath) -> FsFuture<()> {
        Box::pin(async move {
            info!("rename");
            let (from_meta, from) = self
                .resolve_path(from)
                .map_err(|_| FsError::GeneralFailure)
                .await?;
            let (to_meta, to) = self
                .resolve_path(to)
                .map_err(|_| FsError::GeneralFailure)
                .await?;

            match (from_meta.exists, to_meta.exists) {
                (true, false) if !from_meta.is_mount_point => {
                    tokio::fs::rename(from, to)
                        .await
                        .map_err(|_| FsError::GeneralFailure)?;
                    Ok(())
                }
                (true, false) => Err(FsError::NotImplemented), //TODO rename dst in db
                (false, _) => Err(FsError::NotFound),
                (true, true) => Err(FsError::Exists),
            }
        })
    }

    fn copy<'a>(&'a self, from: &'a DavPath, to: &'a DavPath) -> FsFuture<()> {
        Box::pin(async move {
            info!("copy");
            let (from_meta, from) = self
                .resolve_path(from)
                .map_err(|_| FsError::GeneralFailure)
                .await?;
            let (to_meta, to) = self
                .resolve_path(to)
                .map_err(|_| FsError::GeneralFailure)
                .await?;
            match (from_meta.exists, to_meta.exists) {
                (true, false) => {
                    tokio::fs::copy(from, to)
                        .await
                        .map_err(|_| FsError::GeneralFailure)?;
                }
                (false, _) => return Err(FsError::NotFound),
                (true, true) => return Err(FsError::Exists),
            }
            Ok(())
        })
    }

    fn have_props<'a>(
        &'a self,
        path: &'a DavPath,
    ) -> std::pin::Pin<Box<dyn Future<Output = bool> + Send + 'a>> {
        Box::pin(ready(true))
    }

    fn patch_props<'a>(
        &'a self,
        path: &'a DavPath,
        patch: Vec<(bool, DavProp)>,
    ) -> FsFuture<Vec<(StatusCode, DavProp)>> {
        Box::pin(async move {
            error!("NOT IMPLEMENTED patch_props");
            Err(FsError::NotImplemented)
        })
    }

    fn get_props<'a>(&'a self, path: &'a DavPath, do_content: bool) -> FsFuture<Vec<DavProp>> {
        Box::pin(async move {
            error!("NOT IMPLEMENTED get_props");
            Err(FsError::NotImplemented)
        })
    }

    fn get_prop<'a>(&'a self, ppath: &'a DavPath, prop: DavProp) -> FsFuture<Element> {
        fn element(prefix: &str, name: &str, children: Vec<XMLNode>) -> Element {
            Element {
                prefix: Some(prefix.to_string()),
                namespace: None,
                namespaces: None,
                name: name.to_string(),
                attributes: HashMap::default(),
                children,
            }
        }

        Box::pin(async move {
            let (meta, path) = self
                .resolve_path(ppath)
                .map_err(|_| FsError::GeneralFailure)
                .await?;
            //if !self.can_access(&path) {
            //    return Err(FsError::Forbidden);
            //}
            //
            const NC_PREFIX: &str = "nc";
            const OC_PREFIX: &str = "oc";

            match (
                prop.name.as_str(),
                prop.namespace.as_deref().unwrap_or_default(),
            ) {
                ("permissions", DAV_OC_PROP) => {
                    let val = match meta.is_share {
                        true => "SRGDNVCK",
                        false => "RGDNVCK",
                    };
                    return Ok(element(
                        OC_PREFIX,
                        "permissions",
                        vec![XMLNode::Text(val.to_string())],
                    ));
                }
                ("rich-workspace", DAV_NC_PROP) if false => {
                    return Ok(element(
                        NC_PREFIX,
                        "rich-workspace",
                        vec![XMLNode::Text(
                            "# Hello World\nLorem Ipsum, si dolor amet.".to_string(),
                        )],
                    ))
                }
                ("has-preview", DAV_NC_PROP) => {
                    return Ok(element(
                        NC_PREFIX,
                        "has-preview",
                        vec![XMLNode::Text("false".to_string())],
                    ))
                }
                ("mount-type", DAV_NC_PROP) => {
                    let val = match meta.is_share {
                        true => "shared",
                        false => "",
                    };
                    return Ok(element(
                        NC_PREFIX,
                        "mount-type",
                        vec![XMLNode::Text(val.to_string())],
                    ));
                }
                ("sharees", DAV_NC_PROP) if false => {
                    return Ok(element(
                        NC_PREFIX,
                        "sharees",
                        vec![XMLNode::Element(element(
                            "sharee",
                            "nc",
                            vec![
                                XMLNode::Element(element(
                                    "id",
                                    "nc",
                                    vec![XMLNode::Text("1".to_string())],
                                )),
                                XMLNode::Element(element(
                                    "display-name",
                                    "nc",
                                    vec![XMLNode::Text("testuser".to_string())],
                                )),
                                XMLNode::Element(element(
                                    "type",
                                    "nc",
                                    vec![XMLNode::Text("0".to_string())],
                                )),
                            ],
                        ))],
                    ));
                }
                ("is-encrypted", DAV_NC_PROP) => {
                    return Ok(element(
                        NC_PREFIX,
                        "is-encrypted",
                        vec![XMLNode::Text("false".to_string())],
                    ))
                }
                _ => (),
            }
            Err(FsError::NotFound)
        })
    }

    fn get_quota(&self) -> FsFuture<(u64, Option<u64>)> {
        info!("quota");
        Box::pin(async move {
            error!("NOT IMPLEMENTED get_quota");
            Err(FsError::NotImplemented)
        })
    }
}

impl FilesFs {
    async fn fill_cache(&self) -> Result<(), Error> {
        if self.incoming_share_cache.read().await.is_empty() {
            let res = query!(
                "SELECT src_user, src_path, dst_path FROM user_shares WHERE dst_user = ?",
                self.user.0
            )
            .fetch_all(&self.db)
            .await?;

            let mut cache = self.incoming_share_cache.write().await;
            res.into_iter().for_each(|x| {
                cache.insert(
                    PathBuf::from(x.dst_path),
                    (FsUser(x.src_user), PathBuf::from(x.src_path)),
                );
            });
        }

        Ok(())
    }

    async fn shares(&self, rel_root: &DavPath) -> Result<Vec<Box<dyn DavDirEntry>>, Error> {
        let rel_dst_home = rel_root.as_pathbuf();
        let rel_dst_home = rel_dst_home
            .strip_prefix("/")
            .map(|x| x.to_owned())
            .unwrap_or_else(|_| rel_dst_home);

        self.fill_cache().await?;

        let cache = self.incoming_share_cache.read().await;

        let mut cursor = cache.lower_bound(std::ops::Bound::Included(&rel_dst_home));

        let mut share_dir_entries: Vec<Box<dyn DavDirEntry>> = vec![];

        while let Some((dst_path, (src_user, src_path))) =
            cursor
                .key_value()
                .and_then(|x| match x.0.parent() == Some(&rel_dst_home) {
                    true => Some(x),
                    false => None,
                })
        {
            let name = dst_path.strip_prefix(&rel_dst_home).unwrap();

            let mut p = PathBuf::new();
            p.push(src_user.home_dir());
            p.push(src_path);

            let meta = metadata(&p).await.unwrap();

            share_dir_entries.push(Box::new(OxiDirEntry {
                name: name.to_str().unwrap_or_default().into(),
                len: meta.len(),
                dir: p.is_dir(),
                etag: etag(&p, &meta),
                modified: meta.modified().ok(),
                created: meta.created().ok(),
            }));

            cursor.move_next();
        }

        Ok(share_dir_entries)
    }

    async fn resolve_path(&self, rel_root: &DavPath) -> Result<(PathMeta, PathBuf), Error> {
        let rel_dst_home = rel_root.as_pathbuf();
        let rel_dst_home = rel_dst_home
            .strip_prefix("/")
            .map(|x| x.to_owned())
            .unwrap_or_else(|_| rel_dst_home);

        self.fill_cache().await?;

        let cache = self.incoming_share_cache.read().await;

        let cursor = cache.lower_bound(std::ops::Bound::Included(&rel_dst_home));
        let share = cursor
            .key_value()
            .and_then(|x| rel_dst_home.strip_prefix(x.0).map(|y| (y, x.1)).ok());

        if let Some((rel_mnt, (src_user, src_path))) = share {
            let mut p = PathBuf::new();
            p.push(src_user.home_dir());
            p.push(src_path);
            p.push(rel_mnt);

            Ok((
                PathMeta {
                    is_share: true,
                    is_mount_point: rel_mnt == rel_dst_home,
                    owner: *src_user,
                    exists: p.exists(),
                },
                p,
            ))
        } else {
            let mut p = PathBuf::new();
            p.push(&self.user.home_dir());
            p.push(rel_dst_home.clone());

            Ok((
                PathMeta {
                    is_share: false,
                    is_mount_point: false,
                    owner: self.user,
                    exists: p.exists(),
                },
                p,
            ))
        }
    }
}

#[derive(Debug)]
pub struct PathMeta {
    is_share: bool,
    is_mount_point: bool,
    owner: FsUser,
    exists: bool,
}

pub struct DirEntryStream<'a> {
    entry_stream: ReadDir,
    entry_stream_completed: bool,
    intermediate: Vec<BoxFuture<'a, OxiDirEntry>>,
}

impl<'a> DirEntryStream<'a> {
    pub async fn read_dir(path: &Path) -> DirEntryStream<'a> {
        Self {
            entry_stream: read_dir(path).await.unwrap(),
            entry_stream_completed: false,
            intermediate: vec![],
        }
    }
}

impl<'a> Stream for DirEntryStream<'a> {
    type Item = OxiDirEntry;

    fn poll_next(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Option<Self::Item>> {
        let entry_stream_item = match self.entry_stream_completed {
            false => self.entry_stream.poll_next_entry(cx),
            true => Poll::Ready(Ok(None)),
        };
        match entry_stream_item {
            Poll::Ready(Ok(Some(entry))) => self.intermediate.push(Box::pin(async move {
                let meta = entry.metadata().await.unwrap();
                OxiDirEntry {
                    name: entry.file_name().to_str().unwrap_or_default().into(),
                    len: meta.len(),
                    dir: entry.file_type().await.unwrap().is_dir(),
                    etag: etag(&entry.path(), &meta),
                    modified: meta.modified().ok(),
                    created: meta.created().ok(),
                }
            })),
            Poll::Ready(Ok(None)) => {
                self.entry_stream_completed = true;
            }
            _ => (),
        }
        for (i, intermediate) in self.intermediate.iter_mut().enumerate() {
            if let Poll::Ready(x) = intermediate.poll_unpin(cx) {
                let intermediate_fut = self.intermediate.remove(i);
                drop(intermediate_fut);
                return Poll::Ready(Some(x));
            }
        }
        match self.intermediate.is_empty() {
            true => Poll::Ready(None),
            false => Poll::Pending,
        }
    }
}

#[derive(Clone, Debug)]
pub struct OxiDirEntry {
    name: Box<str>,
    len: u64,
    dir: bool,
    etag: String,
    modified: Option<SystemTime>,
    created: Option<SystemTime>,
}

impl DavDirEntry for OxiDirEntry {
    fn name(&self) -> Vec<u8> {
        self.name.as_bytes().to_vec()
    }

    fn metadata(&self) -> FsFuture<Box<dyn DavMetaData>> {
        Box::pin(async move {
            let meta: Box<dyn DavMetaData> = Box::new(self.clone());
            Ok(meta)
        })
    }
}

impl DavMetaData for OxiDirEntry {
    fn len(&self) -> u64 {
        self.len
    }

    fn modified(&self) -> FsResult<std::time::SystemTime> {
        self.modified.ok_or(FsError::GeneralFailure)
    }

    fn created(&self) -> FsResult<std::time::SystemTime> {
        self.created.ok_or(FsError::GeneralFailure)
    }

    fn is_dir(&self) -> bool {
        self.dir
    }

    fn etag(&self) -> Option<String> {
        Some(self.etag.to_string())
    }
}

#[derive(Debug)]
pub struct OxiFile {
    path: PathBuf,
    file: Option<File>,
}

impl OxiFile {
    pub async fn new(path: PathBuf, options: &tokio::fs::OpenOptions) -> FsResult<Self> {
        let file = options
            .open(&path)
            .await
            .map_err(|_| FsError::GeneralFailure)?;
        Ok(Self {
            path,
            file: Some(file),
        })
    }
}

impl DavFile for OxiFile {
    fn metadata(&mut self) -> FsFuture<Box<dyn DavMetaData>> {
        Box::pin(async move {
            let meta = metadata(&self.path).await.unwrap();

            let etag = etag(&self.path, &meta);

            let oxi_meta = OxiFileMeta {
                len: meta.len(),
                is_dir: meta.is_dir(),
                etag,
            };
            let oxi_meta: Box<dyn DavMetaData> = Box::new(oxi_meta);
            Ok(oxi_meta)
        })
    }

    fn write_bytes(&mut self, buf: Bytes) -> FsFuture<()> {
        async move {
            let mut file = self.file.take().unwrap();
            let res = file.write_all(&buf).await;
            self.file = Some(file);
            res.map_err(|e| e.into())
        }
        .boxed()
    }

    fn write_buf(&mut self, mut buf: Box<dyn Buf + Send>) -> FsFuture<()> {
        async move {
            let mut file = self.file.take().unwrap();
            while buf.remaining() > 0 {
                match file.write(buf.chunk()).await {
                    Ok(n) => buf.advance(n),
                    Err(e) => {
                        self.file = Some(file);
                        return Err(e.into());
                    }
                }
            }
            self.file = Some(file);
            Ok(())
        }
        .boxed()
    }

    fn read_bytes(&mut self, count: usize) -> FsFuture<Bytes> {
        async move {
            let mut file = self.file.take().unwrap();
            let mut buf = BytesMut::with_capacity(count);
            let res = unsafe {
                buf.set_len(count);
                file.read(&mut buf).await.map(|n| {
                    buf.set_len(n);
                    buf.freeze()
                })
            };
            self.file = Some(file);
            res.map_err(|e| e.into())
        }
        .boxed()
    }

    fn seek(&mut self, pos: SeekFrom) -> FsFuture<u64> {
        async move {
            let mut file = self.file.take().unwrap();
            let res = file.seek(pos).await;
            self.file = Some(file);
            res.map_err(|e| e.into())
        }
        .boxed()
    }

    fn flush(&mut self) -> FsFuture<()> {
        async move {
            let mut file = self.file.take().unwrap();
            let res = file.flush().await;
            self.file = Some(file);
            res.map_err(|e| e.into())
        }
        .boxed()
    }
}

#[derive(Clone, Debug)]
pub struct OxiFileMeta {
    len: u64,
    is_dir: bool,
    etag: String,
}

impl DavMetaData for OxiFileMeta {
    fn len(&self) -> u64 {
        self.len
    }

    fn modified(&self) -> FsResult<SystemTime> {
        Err(FsError::NotImplemented)
    }

    fn is_dir(&self) -> bool {
        self.is_dir
    }

    fn etag(&self) -> Option<String> {
        Some(self.etag.to_string())
    }
}

fn etag(path: &Path, meta: &Metadata) -> String {
    let modified = meta
        .modified()
        .ok()
        .and_then(|x| x.duration_since(UNIX_EPOCH).ok())
        .map(|x| x.as_secs().to_ne_bytes())
        .unwrap_or_default()
        .to_vec();
    let path = path.to_str().unwrap_or_default().as_bytes();
    let hash: [u8; 16] = md5::compute([path, &modified].concat()).into();
    base64std.encode(hash)
}
