use axum::{http::StatusCode, response::IntoResponse, Json};
use log::error;
use serde::Serialize;
use thiserror::Error;

#[derive(Debug, Error)]
pub enum Error {
    #[error("internal server error")]
    InternalServerError,

    #[error("bad request")]
    BadRequest,

    #[error("csrf check failed")]
    CsrfCheckFailed,

    #[error("sqlx error: {0:?}")]
    Sqlx(#[from] sqlx::Error),

    #[error("base64 error: {0:?}")]
    Base64User(base64::DecodeError),
}

impl IntoResponse for Error {
    fn into_response(self) -> axum::response::Response {
        error!("{:?}", self);
        match self {
            Self::CsrfCheckFailed => (
                StatusCode::PRECONDITION_FAILED,
                Json(JsonError {
                    message: "CSRF check failed".into(),
                }),
            )
                .into_response(),
            Self::Sqlx(_) => StatusCode::INTERNAL_SERVER_ERROR.into_response(),
            Self::InternalServerError => StatusCode::INTERNAL_SERVER_ERROR.into_response(),
            Self::Base64User(_) => StatusCode::BAD_REQUEST.into_response(),
            Self::BadRequest => StatusCode::BAD_REQUEST.into_response(),
        }
    }
}

#[derive(Serialize)]
struct JsonError {
    message: Box<str>,
}
