#![feature(btree_cursors)]

use std::{env, path::PathBuf, sync::Arc};

use axum::{
    async_trait,
    body::Body,
    debug_handler,
    error_handling::HandleErrorLayer,
    extract::{FromRequestParts, Path, Query, Request, State},
    http::{
        header::{AUTHORIZATION, CONTENT_TYPE, USER_AGENT},
        request::Parts,
        uri::PathAndQuery,
        HeaderMap, StatusCode, Uri,
    },
    response::{IntoResponse, Redirect, Response},
    routing::{any, get, head},
    BoxError, Json, Router,
};
use axum_oidc::{
    error::MiddlewareError, EmptyAdditionalClaims, OidcAuthLayer, OidcClaims, OidcLoginLayer,
};
use base64::{engine::general_purpose::STANDARD as base64std, Engine};
use bytes::Bytes;
use dav_fs::FilesFs;
use dotenvy::dotenv;
use fs::FsUser;
use log::debug;
use ocs::{OcsJson, OcsXml};
use rand::{distributions, Rng};
use serde::{Deserialize, Serialize};
use sha3::{Digest, Sha3_256};
use sqlx::{query, query_as, FromRow, MySqlPool};
use tokio::{fs::File, io::AsyncReadExt, net::TcpListener};
use tower::ServiceBuilder;
use tower_http::trace::TraceLayer;
use tower_sessions::{cookie::SameSite, MemoryStore, SessionManagerLayer};
use webdav_handler::{memfs::MemFs, memls::MemLs, DavConfig, DavHandler};

use crate::error::Error;

mod dav_fs;
mod error;
mod fs;
mod ocs;
mod upload_fs;

type HResult<T> = Result<T, Error>;

#[derive(Clone)]
pub struct AppState {
    dav_server: Arc<DavHandler>,
    db: MySqlPool,
    share_password_salt: Arc<str>,
}

#[tokio::main]
async fn main() {
    dotenv().ok();
    env_logger::init();

    let database_url = env::var("DATABASE_URL").expect("DATABASE_URL");

    let database_pool = MySqlPool::connect(&database_url)
        .await
        .expect("working database connection");

    let application_base = env::var("APPLICATION_BASE").expect("APPLICATION_BASE");
    let application_base =
        Uri::from_maybe_shared(application_base).expect("valid APPLICATION_BASE");
    let issuer = env::var("ISSUER").expect("ISSUER");
    let client_id = env::var("CLIENT_ID").expect("CLIENT_ID");
    let client_secret = env::var("CLIENT_SECRET").ok();
    let scopes: Vec<String> = env::var("SCOPES")
        .expect("SCOPES")
        .split(' ')
        .map(String::from)
        .collect::<Vec<_>>();

    let session_store = MemoryStore::default();
    let session_service = ServiceBuilder::new()
        .layer(HandleErrorLayer::new(|_: BoxError| async {
            StatusCode::BAD_REQUEST
        }))
        .layer(SessionManagerLayer::new(session_store).with_same_site(SameSite::Lax));

    let oidc_login_service = ServiceBuilder::new()
        .layer(HandleErrorLayer::new(|e: MiddlewareError| async {
            debug!("auth layer error {:?}", e);
            e.into_response(); //TODO return this response
            StatusCode::INTERNAL_SERVER_ERROR
        }))
        .layer(OidcLoginLayer::<EmptyAdditionalClaims>::new());

    let oidc_auth_service = ServiceBuilder::new()
        .layer(HandleErrorLayer::new(|e: MiddlewareError| async {
            debug!("auth layer error {:?}", e);
            e.into_response(); //TODO return this response
            StatusCode::INTERNAL_SERVER_ERROR
        }))
        .layer(
            OidcAuthLayer::<EmptyAdditionalClaims>::discover_client(
                application_base,
                issuer,
                client_id,
                client_secret,
                scopes,
            )
            .await
            .unwrap(),
        );

    let state = AppState {
        dav_server: Arc::new(
            DavHandler::builder()
                .filesystem(MemFs::new())
                .locksystem(MemLs::new())
                .build_handler(),
        ),
        db: database_pool,
        share_password_salt: "".into(),
    };

    let app = Router::new()
        .route("/index.php/login/flow", get(login_flow))
        .layer(oidc_login_service)
        .route("/index.php/204", get(connectivity_check))
        .route("/status.php", get(status))
        //.route("/remote.php/dav", any(remote_dav))
        .route(
            "/remote.php/dav",
            head(|| async { StatusCode::OK.into_response() }),
        )
        .nest_service(
            "/remote.php/dav/files/:user_id",
            any(user_dav).with_state(state.clone()),
        )
        .nest_service(
            "/remote.php/dav/upload/:user_id",
            any(upload_dav).with_state(state.clone()),
        )
        .nest("/ocs/v2.php", ocs::router(state.clone()))
        .route("/index.php/avatar/:user_id/512", get(avatar_512))
        .route("/index.php/avatar/:user_id/:size", get(|| async { "" }))
        .layer(oidc_auth_service)
        .layer(session_service)
        .layer(TraceLayer::new_for_http())
        .with_state(state);

    let listener = TcpListener::bind("[::]:8080").await.expect("valid address");
    axum::serve(listener, app).await.unwrap()
}

async fn connectivity_check() -> impl IntoResponse {
    StatusCode::NO_CONTENT
}

async fn avatar_512() -> Response {
    let mut buf = vec![];
    let mut image_file = File::open("static/ferris.png").await.unwrap();
    image_file.read_to_end(&mut buf).await;
    let buf = Bytes::from(buf);
    Response::builder()
        .header(CONTENT_TYPE, "image/png")
        .body(Body::from(buf))
        .unwrap()
}

#[derive(Deserialize)]
struct Format {
    format: Option<String>,
}

async fn ocs_default(Query(query): Query<Format>) -> Response {
    if query.format.as_deref() == Some("json") {
        OcsJson::not_implemented(()).into_response()
    } else {
        OcsXml::not_implemented(()).into_response()
    }
}

#[derive(Serialize)]
#[serde(rename_all = "camelCase")]
struct Status {
    installed: bool,
    maintenance: bool,
    needs_db_upgrade: bool,
    version: Box<str>,
    version_string: Box<str>,
    edition: Box<str>,
    productname: Box<str>,
    extended_support: bool,
}

async fn status() -> Json<Status> {
    debug!("GET /status.php");
    Json(Status {
        installed: true,
        maintenance: false,
        needs_db_upgrade: false,
        version: "27.1.3".into(),
        version_string: "27.1.3".into(),
        edition: "".into(),
        productname: "pfzettos Nextcloud Backend".into(),
        extended_support: false,
    })
}

async fn user_dav(
    State(state): State<AppState>,
    user: User,
    Path(user_id): Path<u32>,
    mut req: Request,
) -> Response {
    if user_id != user.id {
        return StatusCode::UNAUTHORIZED.into_response();
    }

    let user_prefix = format!("/remote.php/dav/files/{}", user.id);

    let path = match req.uri().path_and_query() {
        Some(x) => format!(
            "{}{}?{}",
            user_prefix,
            x.path(),
            x.query().unwrap_or_default()
        ),
        None => format!("{}{}", user_prefix, req.uri().path()),
    }
    .parse()
    .unwrap();

    let mut parts = req.uri().clone().into_parts();
    parts.path_and_query = Some(path);
    *req.uri_mut() = Uri::from_parts(parts).unwrap();

    let dav_config = DavConfig::new()
        .strip_prefix(user_prefix)
        .autoindex(true, None)
        .filesystem(Box::new(FilesFs {
            user: FsUser(user.id),
            db: state.db.clone(),
            incoming_share_cache: Arc::default(),
        }));

    state
        .dav_server
        .handle_with(dav_config, req)
        .await
        .into_response()
}

async fn upload_dav(
    State(state): State<AppState>,
    user: User,
    Path(user_id): Path<u32>,
    req: Request,
) -> Response {
    if user_id != user.id {
        return StatusCode::UNAUTHORIZED.into_response();
    }

    let dav_config = DavConfig::new()
        .autoindex(true, None)
        .filesystem(MemFs::new());

    state
        .dav_server
        .handle_with(dav_config, req)
        .await
        .into_response()
}

async fn login_flow(
    State(state): State<AppState>,
    headers: HeaderMap,
    OidcClaims(claims): OidcClaims<EmptyAdditionalClaims>,
) -> HResult<Redirect> {
    let user_id = query!(
        "SELECT id FROM users WHERE oidc_id = ?",
        claims.subject().to_string()
    )
    .fetch_optional(&state.db)
    .await?;

    let user_id = match user_id {
        Some(row) => row.id,
        None => {
            let mut transaction = state.db.begin().await?;
            query!(
                "INSERT INTO users (oidc_id, name) VALUES (?, ?)",
                claims.subject().to_string(),
                claims
                    .preferred_username()
                    .map(|x| x.to_string())
                    .unwrap_or_default()
            )
            .execute(&mut *transaction)
            .await?;

            let id = query!("SELECT LAST_INSERT_ID() AS id")
                .fetch_one(&mut *transaction)
                .await?
                .id as u32;
            transaction.commit().await?;
            id
        }
    };

    let user_token = rand::thread_rng()
        .sample_iter(distributions::Alphanumeric)
        .take(64)
        .map(char::from)
        .collect::<String>();
    let user_token_hash = {
        let mut token_hasher = Sha3_256::default();
        token_hasher.update(&user_token);
        token_hasher.finalize().to_vec()
    };

    let user_agent = headers
        .get(USER_AGENT)
        .and_then(|x| x.to_str().ok())
        .unwrap_or("");

    query!(
        "INSERT INTO user_tokens (user_id, name, hash) VALUES (?, ?, ?)",
        user_id,
        user_agent,
        user_token_hash
    )
    .execute(&state.db)
    .await?;

    let mut transaction = state.db.begin().await?;
    query!(
        "INSERT INTO user_tokens (user_id, name, hash) VALUES (?, ?, ?)",
        user_id,
        user_agent,
        user_token_hash
    )
    .execute(&mut *transaction)
    .await?;

    let token_id = query!("SELECT LAST_INSERT_ID() AS id")
        .fetch_one(&mut *transaction)
        .await?
        .id as u32;
    transaction.commit().await?;

    Ok(Redirect::temporary(&format!(
        "nc://login/server:http://10.50.10.2:8080&user:{}&password:{}",
        token_id, user_token
    )))
}

impl AsRef<MySqlPool> for AppState {
    fn as_ref(&self) -> &MySqlPool {
        &self.db
    }
}

#[derive(FromRow, Clone)]
pub struct User {
    id: u32,
    oidc_id: Box<str>,
    name: Box<str>,
}

impl User {
    pub fn home_dir(&self) -> PathBuf {
        PathBuf::from(format!("./files/{}/", self.id))
    }
    pub fn upload_dir(&self) -> PathBuf {
        PathBuf::from(format!("./upload/{}/", self.id))
    }
}

#[async_trait]
impl<S: AsRef<MySqlPool> + Send + Sync> FromRequestParts<S> for User {
    type Rejection = Result<StatusCode, Error>;

    async fn from_request_parts(parts: &mut Parts, state: &S) -> Result<Self, Self::Rejection> {
        let headers = HeaderMap::from_request_parts(parts, state)
            .await
            .map_err(|_| Err(Error::InternalServerError))?;
        let authorizazion = headers
            .get(AUTHORIZATION)
            .ok_or(Ok(StatusCode::UNAUTHORIZED))?;
        if let Some(basic) = authorizazion
            .to_str()
            .unwrap_or_default()
            .strip_prefix("Basic ")
        {
            let basic = base64std
                .decode(basic)
                .map_err(|e| Err(Error::Base64User(e)))?;
            let basic = std::str::from_utf8(&basic).unwrap_or_default();
            let (username, password) = basic.split_once(':').unwrap_or_default();

            let password_hash = {
                let mut hasher = Sha3_256::default();
                hasher.update(password);
                hasher.finalize().to_vec()
            };

            let user = query_as!(User, "SELECT users.id, users.oidc_id, users.name FROM user_tokens INNER JOIN users ON user_tokens.user_id=users.id WHERE user_tokens.id = ? AND user_tokens.hash = ?", username, password_hash).fetch_optional(state.as_ref()).await.map_err(|x|Err(x.into()))?;

            match user {
                Some(user) => Ok(user),
                None => Err(Ok(StatusCode::UNAUTHORIZED)),
            }
        } else {
            Err(Ok(StatusCode::UNAUTHORIZED))
        }
    }
}
