use axum::{
    http::header::CONTENT_TYPE,
    response::{IntoResponse, Response},
    routing::{get, post},
    Router,
};
use serde::Serialize;

use crate::AppState;

pub mod core;
pub mod files_sharing;
pub mod provisioning_api;

pub fn router(state: AppState) -> Router<AppState> {
    Router::new()
        .route("/cloud/capabilities", get(core::get_status))
        .route("/cloud/users", get(provisioning_api::get_users))
        .route("/cloud/user", get(provisioning_api::get_user))
        .route(
            "/apps/files_sharing/api/v1/shares",
            get(files_sharing::share::get_shares).post(files_sharing::share::create),
        )
        .route(
            "/apps/files_sharing/api/v1/sharees",
            get(files_sharing::sharees::search),
        )
        .with_state(state)
}

#[derive(Serialize)]
pub struct OcsJson<T: Serialize> {
    ocs: Ocs<T>,
}

#[derive(Serialize)]
pub struct OcsXml<T: Serialize>(Ocs<T>);

#[derive(Serialize)]
#[serde(rename = "ocs")]
pub struct Ocs<T: Serialize> {
    pub meta: OcsMeta,
    pub data: T,
}

#[derive(Serialize)]
pub struct OcsMeta {
    pub status: Box<str>,
    pub statuscode: u16,
    pub message: Box<str>,
    pub totalitems: Option<Box<str>>,
    pub itemsperpage: Option<Box<str>>,
}

impl<T: Serialize> OcsJson<T> {
    pub fn ok(data: T) -> Self {
        Self {
            ocs: Ocs {
                meta: OcsMeta {
                    status: "ok".into(),
                    statuscode: 200,
                    message: "ok".into(),
                    totalitems: None,
                    itemsperpage: None,
                },
                data,
            },
        }
    }
    pub fn not_found(data: T) -> Self {
        Self {
            ocs: Ocs {
                meta: OcsMeta {
                    status: "not found".into(),
                    statuscode: 200,
                    message: "not found".into(),
                    totalitems: None,
                    itemsperpage: None,
                },
                data,
            },
        }
    }
    pub fn not_implemented(data: T) -> Self {
        Self {
            ocs: Ocs {
                meta: OcsMeta {
                    status: "not found".into(),
                    statuscode: 200,
                    message: "not found".into(),
                    totalitems: None,
                    itemsperpage: None,
                },
                data,
            },
        }
    }
}

impl<T: Serialize> OcsXml<T> {
    pub fn ok(data: T) -> Self {
        Self(Ocs {
            meta: OcsMeta {
                status: "ok".into(),
                statuscode: 200,
                message: "ok".into(),
                totalitems: None,
                itemsperpage: None,
            },
            data,
        })
    }
    pub fn not_found(data: T) -> Self {
        Self(Ocs {
            meta: OcsMeta {
                status: "not found".into(),
                statuscode: 200,
                message: "not found".into(),
                totalitems: None,
                itemsperpage: None,
            },
            data,
        })
    }
    pub fn not_implemented(data: T) -> Self {
        Self(Ocs {
            meta: OcsMeta {
                status: "not implemented".into(),
                statuscode: 501,
                message: "not implemented".into(),
                totalitems: None,
                itemsperpage: None,
            },
            data,
        })
    }
}

impl<T: Serialize> IntoResponse for OcsJson<T> {
    fn into_response(self) -> axum::response::Response {
        let body = serde_json::to_string(&self).unwrap();
        Response::builder()
            .header(CONTENT_TYPE, "application/json")
            .body(body)
            .unwrap()
            .into_response()
    }
}

impl<T: Serialize> IntoResponse for OcsXml<T> {
    fn into_response(self) -> axum::response::Response {
        let body = serde_xml_rs::to_string(&self).unwrap();
        Response::builder()
            .header(CONTENT_TYPE, "application/xml; charset=utf-8")
            .body(body)
            .unwrap()
            .into_response()
    }
}
