use axum::http::HeaderMap;
use serde::Serialize;

use crate::{error::Error, HResult};

use super::OcsJson;

// https://docs.nextcloud.com/server/latest/developer_manual/_static/openapi.html#/operations/core-ocs-get-capabilities
pub async fn get_status(headers: HeaderMap) -> HResult<OcsJson<CapabilitiesData>> {
    if let Some(Ok("true")) = headers.get("OCS-ApiRequest").map(|x| x.to_str()) {
        Ok(OcsJson::ok(CapabilitiesData {
            version: CapabilitiesVersion {
                major: 0,
                minor: 1,
                micro: 0,
                string: "0.1.0".into(),
                edition: "".into(),
                extended_support: false,
            },
            capabilities: Capabilities {
                dav: DavCapabilities {
                    bulkupload: "1.0".into(),
                    chunking: "1.0".into(),
                },
                user_status: UserStatusCapabilities::default(),
                weather_status: WeatherStatusCapabilities::default(),
                files: FilesCapabilities::default(),
                files_sharing: FilesSharingCapabilities {
                    api_enabled: true,
                    default_permissions: 31,
                    group_sharing: false,
                    public: FilesSharingPublicCapabilities {
                        enabled: true,
                        multiple_links: true,
                        send_mail: false,
                        upload: true,
                        upload_files_drop: true,
                        ..Default::default()
                    },
                    resharing: true,
                    ..Default::default()
                },
            },
        }))
    } else {
        Err(Error::CsrfCheckFailed)
    }
}

#[derive(Serialize)]
pub struct CapabilitiesData {
    version: CapabilitiesVersion,
    capabilities: Capabilities,
}

#[derive(Serialize)]
#[serde(rename_all = "camelCase")]
pub struct CapabilitiesVersion {
    major: i64,
    minor: i64,
    micro: i64,
    string: Box<str>,
    edition: Box<str>,
    extended_support: bool,
}

#[derive(Serialize)]
pub struct Capabilities {
    dav: DavCapabilities,
    user_status: UserStatusCapabilities,
    weather_status: WeatherStatusCapabilities,
    files: FilesCapabilities,
    files_sharing: FilesSharingCapabilities,
}

#[derive(Serialize)]
pub struct DavCapabilities {
    bulkupload: Box<str>,
    chunking: Box<str>,
}

#[derive(Serialize, Default)]
pub struct UserStatusCapabilities {
    enabled: bool,
    restore: bool,
    supports_emoji: bool,
}

#[derive(Serialize, Default)]
pub struct WeatherStatusCapabilities {
    enabled: bool,
}

#[derive(Serialize)]
pub struct FilesCapabilities {
    bigfilechunking: bool,
    blacklisted_files: Box<[Box<str>]>,
    comments: bool,
    undelete: bool,
    version_deletion: bool,
    version_labeling: bool,
    versioning: bool,
}

impl Default for FilesCapabilities {
    fn default() -> Self {
        Self {
            bigfilechunking: true,
            blacklisted_files: Default::default(),
            comments: false,
            undelete: false,
            version_deletion: false,
            version_labeling: false,
            versioning: false,
        }
    }
}

#[derive(Serialize, Default)]
pub struct FilesSharingCapabilities {
    api_enabled: bool,
    default_permissions: u32,
    federation: FilesSharingFederationCapabilities,
    group_sharing: bool,
    public: FilesSharingPublicCapabilities,
    resharing: bool,
    sharebymail: FilesSharingSharebymailCapabilities,
    sharee: FilesSharingShareeCapabilities,
    user: FilesSharingUserCapabilities,
}
#[derive(Serialize, Default)]
pub struct FilesSharingFederationCapabilities {
    expire_date: Expire,
    expire_date_supported: Expire,
    incoming: bool,
    outgoing: bool,
}
#[derive(Serialize, Default)]
pub struct FilesSharingPublicCapabilities {
    enabled: bool,
    expire_date: Expire,
    expire_date_internal: Expire,
    expire_date_remote: Expire,

    multiple_links: bool,

    send_mail: bool,
    upload: bool,
    upload_files_drop: bool,
}
#[derive(Serialize, Default)]
pub struct FilesSharingSharebymailCapabilities {
    enabled: bool,
    expire_date: Expire,
    send_passowrd_by_mail: bool,
}
#[derive(Serialize, Default)]
pub struct FilesSharingUserCapabilities {
    expire_date: Expire,
    send_mail: bool,
}
#[derive(Serialize, Default)]
pub struct FilesSharingShareeCapabilities {
    always_show_unique: bool,
    query_lookup_default: bool,
}

#[derive(Serialize, Default)]
pub struct Expire {
    enabled: bool,
}

/*
//TODO: doesn't work
pub async fn get_avatar(Path((user_id, size)): Path<(u32, u32)>) -> impl IntoResponse {
    let image = Reader::open("static/ferris.png").unwrap().decode().unwrap();
    image.resize(size, size, FilterType::Nearest);
    let mut buf = vec![];
    image.write_to(&mut Cursor::new(&mut buf), ImageFormat::Png);

    Response::builder()
        .status(StatusCode::OK)
        .header(CONTENT_TYPE, "image/png")
        .header(
            CONTENT_DISPOSITION,
            format!("inline; filename=\"avatar.{size}.png\""),
        )
        .header("X-NC-IsCustomAvatar", "1")
        .body(Body::from(buf))
        .unwrap()
}
*/
