use serde::{Deserialize, Serialize};

pub mod share;
pub mod sharees;

#[derive(Debug, Default, PartialEq, Eq)]
pub enum ShareType {
    #[default]
    Unknown,
    User,
    Group,
    Usergroup,
    Link,
    Email,
    #[deprecated]
    Contact,
    Remote,
    Circle,
    Guest,
    RemoteGroup,
    Room,
    UserRoom,
    Deck,
    DeckUser,
    Sciencemesh,
}

impl<'de> Deserialize<'de> for ShareType {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        let value = i32::deserialize(deserializer)?;
        Ok(match value {
            0 => Self::User,
            1 => Self::Group,
            2 => Self::Usergroup,
            3 => Self::Link,
            4 => Self::Email,
            5 => Self::Contact,
            6 => Self::Remote,
            7 => Self::Circle,
            8 => Self::Guest,
            9 => Self::RemoteGroup,
            10 => Self::Room,
            11 => Self::UserRoom,
            12 => Self::Deck,
            13 => Self::DeckUser,
            15 => Self::Sciencemesh,
            _ => Self::Unknown,
        })
    }
}

impl Serialize for ShareType {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        let value = match self {
            Self::Unknown => -1,
            Self::User => 0,
            Self::Group => 1,
            Self::Usergroup => 2,
            Self::Link => 3,
            Self::Email => 4,
            Self::Contact => 5,
            Self::Remote => 6,
            Self::Circle => 7,
            Self::Guest => 8,
            Self::RemoteGroup => 9,
            Self::Room => 10,
            Self::UserRoom => 11,
            Self::Deck => 12,
            Self::DeckUser => 13,
            Self::Sciencemesh => 15,
        };
        value.serialize(serializer)
    }
}

impl ShareType {
    pub fn all() -> Box<[Self]> {
        vec![
            Self::User,
            Self::Group,
            Self::Usergroup,
            Self::Link,
            Self::Email,
            Self::Remote,
            Self::Circle,
            Self::Guest,
            Self::RemoteGroup,
            Self::Room,
            Self::UserRoom,
            Self::Deck,
            Self::DeckUser,
            Self::Sciencemesh,
        ]
        .into()
    }
}

#[derive(Debug, PartialEq, Eq, Clone)]
pub enum SharePermission {
    Read = 1,
    Update = 2,
    Create = 4,
    Delete = 8,
    Share = 16,
}

#[derive(Debug, Default, Clone)]
pub struct SharePermissions(pub Box<[SharePermission]>);

impl SharePermissions {
    pub fn rus() -> Self {
        Self(
            vec![
                SharePermission::Read,
                SharePermission::Update,
                SharePermission::Share,
            ]
            .into(),
        )
    }
}

impl From<u8> for SharePermissions {
    fn from(value: u8) -> Self {
        let mut permissions = vec![];
        if value & 1 == 1 {
            permissions.push(SharePermission::Read);
        }
        if value >> 1 & 1 == 1 {
            permissions.push(SharePermission::Update);
        }
        if value >> 2 & 1 == 1 {
            permissions.push(SharePermission::Create);
        }
        if value >> 3 & 1 == 1 {
            permissions.push(SharePermission::Delete);
        }
        if value >> 4 & 1 == 1 {
            permissions.push(SharePermission::Share);
        }

        Self(permissions.into())
    }
}
impl From<&SharePermissions> for u8 {
    fn from(value: &SharePermissions) -> Self {
        let mut out = 0;
        for p in value.0.iter() {
            match p {
                SharePermission::Read => out += 1,
                SharePermission::Update => out += 2,
                SharePermission::Create => out += 4,
                SharePermission::Delete => out += 8,
                SharePermission::Share => out += 16,
            }
        }
        out
    }
}

impl<'de> Deserialize<'de> for SharePermissions {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        let value = u8::deserialize(deserializer)
            .map(|x| x.into())
            .unwrap_or_else(|_| Self::rus());

        Ok(value)
    }
}

impl Serialize for SharePermissions {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        u8::from(self).serialize(serializer)
    }
}
