use axum::{
    extract::State,
    response::{IntoResponse, Response},
    Form,
};
use serde::{Deserialize, Serialize};
use sqlx::query;
use time::{macros::format_description, Time};

use crate::{error::Error, ocs::OcsXml, AppState, HResult, User};

use super::{SharePermission, SharePermissions, ShareType};

#[derive(Deserialize, Debug)]
pub struct CreateShareData {
    attributes: Option<Box<str>>,
    #[serde(default, rename = "expireDate")]
    expire_date: Option<Box<str>>,
    #[serde(default)]
    label: Box<str>,
    #[serde(default)]
    note: Box<str>,
    #[serde(default)]
    password: Box<str>,
    path: Option<Box<str>>,
    #[serde(default = "SharePermissions::rus")]
    permissions: SharePermissions,
    #[serde(default, rename = "publicUpload")]
    public_upload: bool,
    #[serde(default, rename = "shareType")]
    share_type: ShareType,
    #[serde(rename = "shareWith")]
    share_with: Option<Box<str>>,
}

pub async fn create(
    State(state): State<AppState>,
    user: User,
    Form(form): Form<CreateShareData>,
) -> HResult<Response> {
    match form.share_type {
        ShareType::User => Ok(create_user_share(&state, &user, &form)
            .await?
            .into_response()),
        _ => todo!(),
    }
}

async fn create_user_share(
    state: &AppState,
    user: &User,
    data: &CreateShareData,
) -> HResult<OcsXml<UserShare>> {
    // shareWith is required
    let share_with = match &data.share_with {
        Some(x) => x.as_ref(),
        None => return Err(Error::BadRequest),
    };

    // check if user_to is valid
    if query!("SELECT id FROM users WHERE id = ?", share_with)
        .fetch_optional(&state.db)
        .await?
        .is_none()
    {
        return Err(Error::BadRequest);
    }

    //TODO: parse expire_date
    let expires_at: Option<Time> = None;

    //TODO: change to owner of orig if resharing
    let dst_user_id = user.id;
    let dst_user_name = user.name.clone();

    //TODO: check permissions of user on data.path

    let share_id = {
        let mut transaction = state.db.begin().await?;
        //query!("INSERT INTO user_shares (dst_user_id, from_user_id, to_user_id, expires_at, note, path, permissions) VALUES (?, ?, ?, ?, ?, ?, ?)", &dst_user_id, &user.id, &share_with, expires_at, &data.note, &data.path, u8::from(&data.permissions)).execute(&state.db).await?;

        let row = query!("SELECT LAST_INSERT_ID() as id")
            .fetch_one(&mut *transaction)
            .await?;
        transaction.commit().await?;
        row.id
    };

    Ok(OcsXml::ok(UserShare {
        id: share_id,
        share_type: ShareType::User,
        uid_owner: user.id,
        displayname_owner: user.name.clone(),
        permissions: data.permissions.clone(),
        can_edit: data.permissions.0.contains(&SharePermission::Update),
        can_delete: data.permissions.0.contains(&SharePermission::Delete),
        expiration: expires_at.and_then(|x| {
            x.format(format_description!(
                "[year]-[month]-[day] [hour]:[minute]:[second]"
            ))
            .ok()
            .map(|x| x.into())
        }),
        uid_file_owner: dst_user_id,
        note: data.note.clone(),
        label: data.label.clone(),
        displayname_file_owner: dst_user_name,
        ..Default::default()
    }))
}

#[derive(Default, Serialize)]
struct UserShare {
    id: u64,
    share_type: ShareType,
    uid_owner: u32,
    displayname_owner: Box<str>,
    permissions: SharePermissions,
    can_edit: bool,
    can_delete: bool,
    stime: (),
    parent: Box<str>,
    expiration: Option<Box<str>>,
    token: Box<str>,
    uid_file_owner: u32,
    note: Box<str>,
    label: Box<str>,
    displayname_file_owner: Box<str>,
    path: Box<str>,
    item_type: Box<str>,
    item_permissions: u8,
    mimetype: Box<str>,
    has_preview: bool,
    storage_id: Box<str>,
    storage: Box<str>,
    item_source: (),
    file_source: (),
    file_parent: (),
    file_target: (),
    item_size: (),
    item_mtime: (),
    share_with: Box<str>,
    share_with_displayname: Box<str>,
    share_with_displayname_unique: Box<str>,
    status: (),
    mail_send: bool,
    hide_download: bool,
    attributes: (),
    tags: Box<[()]>,
}

pub async fn get_shares() -> HResult<OcsXml<()>> {
    Ok(OcsXml::ok(()))
}
