use axum::{
    extract::{Query, State},
    response::IntoResponse,
};
use serde::{Deserialize, Serialize};
use sqlx::query;

use crate::{
    ocs::{OcsJson, OcsXml},
    AppState, HResult, User,
};

use super::ShareType;

#[derive(Debug, Deserialize)]
pub struct SearchQuery {
    #[serde(rename = "itemType")]
    item_type: Option<Box<str>>,

    #[serde(default)]
    lookup: bool,

    #[serde(default = "default_page")]
    page: u64,

    #[serde(rename = "perPage", default = "default_per_page")]
    per_page: u64,

    search: Box<str>,

    #[serde(default = "default_share_types")]
    share_type: Box<[ShareType]>,
}

fn default_page() -> u64 {
    1
}

fn default_per_page() -> u64 {
    200
}

fn default_share_types() -> Box<[ShareType]> {
    ShareType::all()
}

pub async fn search(
    State(state): State<AppState>,
    user: User,
    Query(query): Query<SearchQuery>,
) -> HResult<OcsJson<SearchResult>> {
    let mut transaction = state.db.begin().await?;
    let (rough_users, exact_users) = if query.share_type.contains(&ShareType::User) {
        let rough = query!(
            "SELECT id, name FROM users WHERE MATCH(name) AGAINST (? IN NATURAL LANGUAGE MODE) AND name != ?",
            &query.search,
            &query.search,
        )
        .fetch_all(&mut *transaction)
        .await?
        .into_iter()
        .map(|x| SearchUser {
            icon: "user-icon".into(),
            label: x.name.clone().into(),
            share_with_displayname_unique: x.name.into(),
            value: SearchUserValue {
                share_type: ShareType::User,
                share_with: x.id.to_string().into(),
            },
            ..Default::default()
        })
        .collect();

        let exact = query!("SELECT id, name FROM users WHERE name = ?", &query.search)
            .fetch_all(&mut *transaction)
            .await?
            .into_iter()
            .map(|x| SearchUser {
                icon: "user-icon".into(),
                label: x.name.clone().into(),
                share_with_displayname_unique: x.name.into(),
                value: SearchUserValue {
                    share_type: ShareType::User,
                    share_with: x.id.to_string().into(),
                },
                ..Default::default()
            })
            .collect();
        (rough, exact)
    } else {
        (vec![], vec![])
    };

    transaction.commit().await?;

    Ok(OcsJson::ok(SearchResult {
        users: rough_users.into(),
        exact: ExactSearchResult {
            users: exact_users.into(),
            ..Default::default()
        },
        ..Default::default()
    }))
}

#[derive(Serialize, Default)]
pub struct SearchResult {
    circles: Box<[()]>,
    emails: Box<[()]>,
    exact: ExactSearchResult,
    groups: Box<[()]>,
    lookup: Box<[()]>,
    #[serde(rename = "lookupEnabled")]
    lookup_enabled: bool,
    remote_groups: Box<[()]>,
    remotes: Box<[()]>,
    rooms: Box<[()]>,
    users: Box<[SearchUser]>,
}

#[derive(Serialize, Default)]
pub struct ExactSearchResult {
    circles: Box<[()]>,
    emails: Box<[()]>,
    groups: Box<[()]>,
    lookup: Box<[()]>,
    remote_groups: Box<[()]>,
    remotes: Box<[()]>,
    rooms: Box<[()]>,
    users: Box<[SearchUser]>,
}

#[derive(Default, Serialize)]
pub struct SearchUser {
    icon: Box<str>,
    label: Box<str>,
    #[serde(rename = "shareWithDisplayNameUnique")]
    share_with_displayname_unique: Box<str>,
    status: SearchUserStatus,
    subline: Box<str>,
    value: SearchUserValue,
}

#[derive(Serialize)]
pub struct SearchUserStatus {
    #[serde(rename = "clearAt")]
    clear_at: (),
    icon: (),
    message: (),
    status: Box<str>,
}

impl Default for SearchUserStatus {
    fn default() -> Self {
        Self {
            clear_at: Default::default(),
            icon: Default::default(),
            message: Default::default(),
            status: "offline".into(),
        }
    }
}

#[derive(Default, Serialize)]
pub struct SearchUserValue {
    #[serde(rename = "shareType")]
    share_type: ShareType,
    #[serde(rename = "shareWith")]
    share_with: Box<str>,
}
